/* 
/ test_cgi.c
/
/ Test cases for dataSeltzer CGI
/
/ Author: Sandro Furieri a.furieri@lqt.it
/
/ Copyright (C) 2014  Alessandro Furieri
/
/   This program is free software: you can redistribute it and/or modify
/   it under the terms of the GNU Affero General Public License as published by
/   the Free Software Foundation, either version 3 of the License, or
/   (at your option) any later version.
/
/   This program is distributed in the hope that it will be useful,
/   but WITHOUT ANY WARRANTY; without even the implied warranty of
/   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/   GNU Affero General Public License for more details.
/
/   You should have received a copy of the GNU Affero General Public License
/   along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>

struct pseudo_url
{
/* a pseudo URL to be tested */
    int test_no;
    char *string;
    struct pseudo_url *next;
};

struct tests
{
/* list of all testcases */
    struct pseudo_url *first;
    struct pseudo_url *last;
    int next_test;
};

static struct tests *
alloc_tests (void)
{
/* allocating an empty list of testcases */
    struct tests *list = malloc (sizeof (struct tests));
    list->first = NULL;
    list->last = NULL;
    list->next_test = 1;
    return list;
}

static void
destroy_tests (struct tests *list)
{
/* memory cleanup - destroyng the list of testcases */
    struct pseudo_url *pu;
    struct pseudo_url *pun;
    if (list == NULL)
	return;
    pu = list->first;
    while (pu != NULL)
      {
	  pun = pu->next;
	  if (pu->string != NULL)
	      free (pu->string);
	  free (pu);
	  pu = pun;
      }
    free (list);
}

static void
insert_test (struct tests *list, const char *string)
{
/* inserting a new testcase into the list */
    struct pseudo_url *p;
    int len;
    if (strncmp (string, "QUERY_STRING=", 13) != 0)
	return;
    len = strlen (string);
    p = malloc (sizeof (struct pseudo_url));
    p->test_no = list->next_test++;
    p->string = malloc (len + 1);
    strcpy (p->string, string);
    p->next = NULL;
    if (list->first == NULL)
	list->first = p;
    if (list->last != NULL)
	list->last->next = p;
    list->last = p;
}

static void
do_parse_tests (struct tests *list)
{
/* attempting to parse the "pseudo_url.tests" files */
    char line[8192];
    char *p = line;
    int c;
    FILE *in = fopen ("./pseudo_url.tests", "r");
    if (in == NULL)
	return;
    while ((c = getc (in)) != EOF)
      {
	  if (c == '\r')
	      continue;
	  if (c == '\n')
	    {
		*p = '\0';
		insert_test (list, line);
		p = line;
		continue;
	    }
	  *p++ = c;
      }
    fclose (in);
}

static int
execute_testcase (FILE * valgrind_script, struct pseudo_url *test)
{
/* executing a single testcase */
    char *envp[2];
    pid_t pid;
    char out_path[1024];
    char err_path[1024];

/* updating the Valgrind script */
    fprintf (valgrind_script, "# test-%d\n", test->test_no);
    fprintf (valgrind_script, "export \"QUERY_STRING=%s\"\n", test->string);
    fprintf (valgrind_script,
	     "valgrind --track-origins=yes --tool=memcheck --num-callers=20 "
	     "--leak-check=full --show-reachable=yes ../dataSeltzer >test%d.html\n",
	     test->test_no);
    fprintf (valgrind_script, "rm -f test%d.html\n\n", test->test_no);

/* setting the QUERY_STRING environment variable */
    envp[0] = test->string;
    envp[1] = NULL;

/* preparing stdout and stderr paths */
    sprintf (out_path, "./test%d.html", test->test_no);
    sprintf (err_path, "./test%d_err", test->test_no);

/* forking a child process */
    pid = fork ();
    if (pid == 0)
      {
	  /* this is the child process ... real test happens here */

	  /* re-directing stout and stderr */
	  if (freopen (out_path, "w", stdout) == NULL)
	    {
		fprintf (stderr, "unable to redirect stdout\n");
		return -1;
	    }
	  if (freopen (err_path, "w", stderr) == NULL)
	    {
		fprintf (stderr, "unable to redirect stderr\n");
		return -2;
	    }
	  /* testing the CGI component */
	  execle ("../dataSeltzer", "dataSeltzer", NULL, envp);
      }
    else
      {
	  /* parent waits for child termination */
	  int status;
	  waitpid (pid, &status, 0);
      }
    unlink (out_path);
    unlink (err_path);
    fprintf (stderr, "test %d - done\n", test->test_no);
    return 0;
}

int
main (void)
{
/* dispatching all testcases */
    int ret = 0;
    FILE *valgrind_script;
    struct tests *list = alloc_tests ();
    struct pseudo_url *test;

/* parsing the "pseudo_url.tests" file */
    do_parse_tests (list);

/* opening the script supporting Valgrind tests */
    valgrind_script = fopen ("./test_under_valgrind.sh", "w");
    if (valgrind_script == NULL)
      {
	  fprintf (stderr,
		   "Unable to open the \"test_under_vargring.sh\" script !!!\n");
	  destroy_tests (list);
	  return -1;
      }

/* executing all testcases */
    test = list->first;
    while (test != NULL)
      {
	  if (execute_testcase (valgrind_script, test) != 0)
	    {
		ret = 0 - test->test_no;
		break;
	    }
	  test = test->next;
      }
    fclose (valgrind_script);

/* final cleanup */
    destroy_tests (list);
    return ret;
}
